@@ -3,6 +3,7 @@
 Copyright (C) 2003, 2004, 2005, 2006, 2008, 2009, 2010
   National Institute of Advanced Industrial Science and Technology (AIST)
   Registration Number H15PRO167
+Copyright (C) 2012, 2013, 2014, 2015  K. Handa  <handa@gnu.org>
 
 This file is part of libotf.
 
@@ -380,7 +381,6 @@ update_seq_area ()
   if (otf)
     {
       char *script_name = NULL, *langsys_name = NULL, buf[10];
-      char *str;
 
       if (script_tag)
 	{
@@ -397,37 +397,41 @@ update_seq_area ()
       OTF_drive_gdef (otf, &gstring);
       if (otf->gsub)
 	{
-	  str = get_features (&otf->gsub->FeatureList, &gsub);
-	  if (str)
+	  char *gsub_features = get_features (&otf->gsub->FeatureList, &gsub);
+
+	  if (gsub_features)
 	    {
 	      if (do_alternate_subst)
 		OTF_drive_gsub_alternate (otf, &gstring,
-					  script_name, langsys_name, str);
+					  script_name, langsys_name,
+					  gsub_features);
 	      else
 		{
-		  OTF_drive_gsub_with_log (otf, &gstring,
-					   script_name, langsys_name, str);
+		  OTF_drive_gsub_features (otf, &gstring,
+					   script_name, langsys_name,
+					   gsub_features);
 		  logsize = gstring.used * 2;
 		  log = alloca (sizeof (OTF_Tag) * logsize);
 		  for (i = 0; i < gstring.used; i++)
 		    {
-		      int idx = gstring.glyphs[i].positioning_type >> 4;
+		      int idx = OTF_POSITIONING_TYPE_GET_FEATURE (gstring.glyphs + i);
 		      if (idx)
 			log[i] = otf->gsub->FeatureList.Feature[idx - 1].FeatureTag;
 		      else
 			log[i] = 0;
 		    }
 		}
-	      free (str);
+	      free (gsub_features);
 	    }
 	}
       if (otf->gpos)
 	{
-	  str = get_features (&otf->gpos->FeatureList, &gpos);
-	  if (str)
+	  char *gpos_features = get_features (&otf->gpos->FeatureList, &gpos);
+	  if (gpos_features)
 	    {
-	      OTF_drive_gpos_with_log (otf, &gstring,
-				       script_name, langsys_name, str);
+	      OTF_drive_gpos_features (otf, &gstring,
+				       script_name, langsys_name,
+				       gpos_features);
 	      if (log)
 		{
 		  if (logsize < gstring.used)
@@ -447,12 +451,11 @@ update_seq_area ()
 		}
 	      for (i = 0; i < gstring.used; i++)
 		{
-		  int idx = gstring.glyphs[i].positioning_type >> 4;
+		  int idx = OTF_POSITIONING_TYPE_GET_FEATURE (gstring.glyphs + i);
 		  if (idx)
 		    log[i] = otf->gpos->FeatureList.Feature[idx - 1].FeatureTag;
-
 		}
-	      free (str);
+	      free (gpos_features);
 	    }
 	}
     }
@@ -494,11 +497,11 @@ update_seq_area ()
 	    continue;
 	  advance = bmp->advance;
 	}
-      if (g->positioning_type & 0xF)
+      if (OTF_POSITIONING_TYPE_GET_FORMAT (g))
 	{
 	  while (1)
 	    {
-	      switch (g->positioning_type & 0xF)
+	      switch (OTF_POSITIONING_TYPE_GET_FORMAT (g))
 		{
 		case 1: case 2:
 		  {
@@ -532,8 +535,29 @@ update_seq_area ()
 		default: 		/* i.e. case 6 */
 		  if (! mark)
 		    break;
-		  prev = mark;
-		  prev_width = 0;
+		  {
+		    int distance = OTF_POSITIONING_TYPE_GET_MARKDISTANCE (g);
+
+		    if (distance > 0) 
+		      {
+			for (prev = g - 1; prev >= gstring.glyphs; prev--) 
+			  {
+			    if (prev->glyph_id > 0)
+			      {
+				distance--;
+				if (distance == 0)
+				  break;
+			      }
+			  }
+			if (prev < gstring.glyphs)
+			  prev = mark;
+		      }
+		    else
+		      {
+			prev = mark;
+		      }
+		    prev_width = 0;
+		  }
 		label_adjust_anchor:
 		  {
 		    int base_x, base_y, mark_x, mark_y;
@@ -553,7 +577,7 @@ update_seq_area ()
 		}
 	      if (i + 1 == gstring.used
 		  || gstring.glyphs[i + 1].glyph_id
-		  || ! (gstring.glyphs[i + 1].positioning_type & 0xF))
+		  || ! (OTF_POSITIONING_TYPE_GET_FORMAT (gstring.glyphs + i + 1)))
 		break;
 	      i++, g++;
 	    }
