@@ -28,6 +28,10 @@ write to the Free Software Foundation, Inc., 59 Temple Place, Suite
 extern "C" {
 #endif
 
+#ifdef HAVE_ALLOCA_H
+#include <alloca.h>
+#endif
+
 #define OTF_ERROR(err, arg)	\
   return (otf__error ((err), errfmt, (arg)), errret)
 
