@@ -1040,8 +1040,6 @@ read_class_def (OTF *otf, OTF_Stream *stream, long offset, OTF_ClassDef *class)
 	= read_glyph_ids (otf, stream,
 			  (OTF_GlyphID **) &class->f.f1.ClassValueArray,
 			  0, -1);
-      if (! class->f.f1.GlyphCount)
-	return -1;
     }
   else if (class->ClassFormat == 2)
     {
@@ -1049,8 +1047,6 @@ read_class_def (OTF *otf, OTF_Stream *stream, long offset, OTF_ClassDef *class)
 	= read_range_records (otf, stream,
 			      (OTF_RangeRecord **)
 			      &class->f.f2.ClassRangeRecord);
-      if (! class->f.f2.ClassRangeCount)
-	return -1;
     }
   else
     OTF_ERROR (OTF_ERROR_TABLE, " (Invalid format)");
@@ -1424,18 +1420,16 @@ read_lookup_list (OTF *otf, OTF_Stream *stream, long offset,
 
 /*** (1-8) Structures common to GSUB and GPOS */
 
-static unsigned
+static int
 read_lookup_record_list (OTF *otf, OTF_Stream *stream,
 			 OTF_LookupRecord **record, int count)
 {
   char *errfmt = "LookupRecord%s";
-  unsigned errret = 0;
+  int errret = -1;
   int i;
 
   if (count < 0)
     READ_UINT16 (stream, count);
-  if (! count)
-    OTF_ERROR (OTF_ERROR_TABLE, " (zero count)");
   OTF_MALLOC (*record, count, "");
   for (i = 0; i < count; i++)
     {
@@ -1476,7 +1470,7 @@ read_rule_list (OTF *otf, OTF_Stream *stream, long offset, OTF_Rule **rule)
 			  (*rule)[i].GlyphCount) < 0)
 	return errret;
       if (read_lookup_record_list (otf, stream, &(*rule)[i].LookupRecord,
-				   (*rule)[i].LookupCount) == 0)
+				   (*rule)[i].LookupCount) < 0)
 	return errret;
     }
   RESTORE_STREAM (stream, state);
@@ -1550,7 +1544,7 @@ read_class_rule_list (OTF *otf, OTF_Stream *stream, long offset,
 			  0, (*rule)[i].GlyphCount - 1) < 0)
 	return errret;
       if (read_lookup_record_list (otf, stream, &(*rule)[i].LookupRecord,
-				   (*rule)[i].LookupCount) == 0)
+				   (*rule)[i].LookupCount) < 0)
 	return errret;
     }
   RESTORE_STREAM (stream, state);
@@ -1606,6 +1600,8 @@ read_chain_rule_list (OTF *otf, OTF_Stream *stream, long offset,
     READ_OFFSET (stream, (*rule)[i].offset);
   for (i = 0; i < count; i++)
     {
+      int lookup_record_count;
+
       SEEK_STREAM (stream, offset + (*rule)[i].offset);
       (*rule)[i].BacktrackGlyphCount
 	= read_glyph_ids (otf, stream, &(*rule)[i].Backtrack, 0, -1);
@@ -1615,11 +1611,12 @@ read_chain_rule_list (OTF *otf, OTF_Stream *stream, long offset,
 	OTF_ERROR (OTF_ERROR_TABLE, " (zero count)");
       (*rule)[i].LookaheadGlyphCount
 	= read_glyph_ids (otf, stream, &(*rule)[i].LookAhead, 0, -1);
-      (*rule)[i].LookupCount
+      lookup_record_count 
 	= read_lookup_record_list (otf, stream,
 				   &(*rule)[i].LookupRecord, -1);
-      if (! (*rule)[i].LookupCount)
+      if (lookup_record_count < 0)
 	return errret;
+      (*rule)[i].LookupCount = lookup_record_count;
     }
   return count;
 }
@@ -1680,6 +1677,8 @@ read_chain_class_rule_list (OTF *otf, OTF_Stream *stream, long offset,
     }
   for (i = 0; i < count; i++)
     {
+      int lookup_record_count;
+
       SEEK_STREAM (stream, offset + (*rule)[i].offset);
       (*rule)[i].BacktrackGlyphCount
 	= read_glyph_ids (otf, stream,
@@ -1692,11 +1691,12 @@ read_chain_class_rule_list (OTF *otf, OTF_Stream *stream, long offset,
       (*rule)[i].LookaheadGlyphCount
 	= read_glyph_ids (otf, stream,
 			  (OTF_GlyphID **) &(*rule)[i].LookAhead, 0, -1);
-      (*rule)[i].LookupCount
+      lookup_record_count
 	= read_lookup_record_list (otf, stream,
 				   &(*rule)[i].LookupRecord, -1);
-      if (! (*rule)[i].LookupCount)
+      if (lookup_record_count < 0)
 	return errret;
+      (*rule)[i].LookupCount = lookup_record_count;
     }
   return count;
 }
@@ -1834,9 +1834,11 @@ read_chain_context3 (OTF *otf, OTF_Stream *stream, long offset,
   count = read_coverage_list (otf, stream, offset,
 			      &chain_context3->LookAhead, -1);
   chain_context3->LookaheadGlyphCount = (unsigned) count;
-  chain_context3->LookupCount
-    = read_lookup_record_list (otf, stream,
-			       &chain_context3->LookupRecord, -1);
+  count = read_lookup_record_list (otf, stream,
+				   &chain_context3->LookupRecord, -1);
+  if (count < 0)
+    return -1;
+  chain_context3->LookupCount = count;
   return 0;
 }
 
@@ -2968,7 +2970,7 @@ OTF_open (const char *otf_name)
 	  && strncasecmp (ext + 1, "ttf", 3)
 	  && strncasecmp (ext + 1, "ttc", 3)))
     OTF_ERROR (OTF_ERROR_FILE, otf_name);
-  fp = fopen (otf_name, "r");
+  fp = fopen (otf_name, "rb");
   if (! fp)
     OTF_ERROR (OTF_ERROR_FILE, otf_name);
   otf = calloc (1, sizeof (OTF));
